@@ -21,7 +21,7 @@
  * Boston, MA 02111-1307, USA.
  */
 
-#include "config.h"
+#include <config.h>
 
 #include <string.h>
 #include <stdlib.h>
@@ -30,31 +30,24 @@
 #include "pango-utils.h"
 #include "pangowin32.h"
 #include "pangowin32-private.h"
-#include "modules.h"
 
-#define PANGO_TYPE_WIN32_FONT            (pango_win32_font_get_type ())
-#define PANGO_WIN32_FONT(object)         (G_TYPE_CHECK_INSTANCE_CAST ((object), PANGO_TYPE_WIN32_FONT, PangoWin32Font))
-#define PANGO_WIN32_FONT_CLASS(klass)    (G_TYPE_CHECK_CLASS_CAST ((klass), PANGO_TYPE_WIN32_FONT, PangoWin32FontClass))
-#define PANGO_WIN32_IS_FONT(object)      (G_TYPE_CHECK_INSTANCE_TYPE ((object), PANGO_TYPE_WIN32_FONT))
-#define PANGO_WIN32_IS_FONT_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), PANGO_TYPE_WIN32_FONT))
-#define PANGO_WIN32_FONT_GET_CLASS(obj)  (G_TYPE_INSTANCE_GET_CLASS ((obj), PANGO_TYPE_WIN32_FONT, PangoWin32FontClass))
-
-#define PANGO_WIN32_UNKNOWN_FLAG 0x10000000
+#define CH_IS_UNIHAN_BMP(ch) ((ch) >= 0x3400 && (ch) <= 0x9FFF)
+#define CH_IS_UNIHAN(ch) (CH_IS_UNIHAN_BMP (ch) || \
+			  ((ch) >= 0x20000 && (ch) <= 0x2A6DF) ||	\
+			  ((ch) >= 0x2F800 && (ch) <= 0x2FA1F))
 
 HDC pango_win32_hdc;
 OSVERSIONINFO pango_win32_os_version_info;
 gboolean pango_win32_debug = FALSE;
 
-typedef struct _PangoWin32FontClass   PangoWin32FontClass;
-
-struct _PangoWin32FontClass
-{
-  PangoFontClass parent_class;
-};
-
 static void pango_win32_font_dispose    (GObject             *object);
 static void pango_win32_font_finalize   (GObject             *object);
 
+static gboolean pango_win32_font_real_select_font        (PangoFont *font,
+							  HDC        hdc);
+static void     pango_win32_font_real_done_font          (PangoFont *font);
+static double   pango_win32_font_real_get_metrics_factor (PangoFont *font);
+
 static PangoFontDescription *pango_win32_font_describe          (PangoFont        *font);
 static PangoCoverage        *pango_win32_font_get_coverage      (PangoFont        *font,
 								 PangoLanguage    *lang);
@@ -70,6 +63,13 @@ static void                  pango_win32_font_get_glyph_extents (PangoFont
 								 PangoRectangle   *logical_rect);
 static PangoFontMetrics *    pango_win32_font_get_metrics       (PangoFont        *font,
 								 PangoLanguage    *lang);
+static PangoFontMap *        pango_win32_font_get_font_map      (PangoFont        *font);
+
+static gboolean pango_win32_font_real_select_font      (PangoFont *font,
+							HDC        hdc);
+static void     pango_win32_font_real_done_font        (PangoFont *font);
+static double   pango_win32_font_real_get_metrics_factor (PangoFont *font);
+
 static HFONT                 pango_win32_get_hfont              (PangoFont        *font);
 static void                  pango_win32_get_item_properties    (PangoItem        *item,
 								 PangoUnderline   *uline,
@@ -78,13 +78,16 @@ static void                  pango_win32_get_item_properties    (PangoItem
 								 PangoAttrColor   *bg_color,
 								 gboolean         *bg_set);
 
-static inline HFONT
+static HFONT
 pango_win32_get_hfont (PangoFont *font)
 {
   PangoWin32Font *win32font = (PangoWin32Font *)font;
   PangoWin32FontCache *cache;
   TEXTMETRIC tm;
 
+  if (!win32font)
+    return NULL;
+
   if (!win32font->hfont)
     {
       cache = pango_win32_font_map_get_font_cache (win32font->fontmap);
@@ -121,16 +124,6 @@ PangoContext *
 pango_win32_get_context (void)
 {
   PangoContext *result;
-  static gboolean registered_modules = FALSE;
-  int i;
-
-  if (!registered_modules)
-    {
-      registered_modules = TRUE;
-      
-      for (i = 0; _pango_included_win32_modules[i].list; i++)
-        pango_module_register (&_pango_included_win32_modules[i]);
-    }
 
   result = pango_context_new ();
   pango_context_set_font_map (result, pango_win32_font_map_for_display ());
@@ -145,6 +138,8 @@ pango_win32_font_init (PangoWin32Font *win32font)
 {
   win32font->size = -1;
 
+  win32font->metrics_by_lang = NULL;
+
   win32font->glyph_info = g_hash_table_new_full (NULL, NULL, NULL, g_free);
 }
 
@@ -185,7 +180,7 @@ pango_win32_get_dc (void)
  *
  * Returns wether debugging is turned on.
  * 
- * Returns: %TRUE if debugging is turned on.
+ * Return value: %TRUE if debugging is turned on.
  * 
  * Since: 1.2
  */
@@ -209,6 +204,11 @@ pango_win32_font_class_init (PangoWin32FontClass *class)
   font_class->find_shaper = pango_win32_font_find_shaper;
   font_class->get_glyph_extents = pango_win32_font_get_glyph_extents;
   font_class->get_metrics = pango_win32_font_get_metrics;
+  font_class->get_font_map = pango_win32_font_get_font_map;
+
+  class->select_font = pango_win32_font_real_select_font;
+  class->done_font = pango_win32_font_real_done_font;
+  class->get_metrics_factor = pango_win32_font_real_get_metrics_factor;
 
   pango_win32_get_dc ();
 }
@@ -267,13 +267,13 @@ pango_win32_render (HDC               hdc,
       PING (("num_glyphs:%d", glyphs->num_glyphs));
       for (i = 0; i < glyphs->num_glyphs; i++)
 	{
-	  printf (" %d:%d", glyphs->glyphs[i].glyph, glyphs->glyphs[i].geometry.width);
+	  g_print (" %d:%d", glyphs->glyphs[i].glyph, glyphs->glyphs[i].geometry.width);
 	  if (glyphs->glyphs[i].geometry.x_offset != 0 || 
 	      glyphs->glyphs[i].geometry.y_offset != 0)
-	    printf (":%d,%d", glyphs->glyphs[i].geometry.x_offset,
-		    glyphs->glyphs[i].geometry.y_offset);
+	    g_print (":%d,%d", glyphs->glyphs[i].geometry.x_offset,
+		     glyphs->glyphs[i].geometry.y_offset);
 	}
-      printf ("\n");
+      g_print ("\n");
     }
 #endif
 
@@ -314,11 +314,12 @@ pango_win32_render (HDC               hdc,
        * point zero (just spacing).
        */
       while (i < glyphs->num_glyphs &&
-	     (glyphs->glyphs[i].glyph == 0 || cur_y_offset == glyphs->glyphs[i].geometry.y_offset))
+	     (glyphs->glyphs[i].glyph == PANGO_GLYPH_EMPTY ||
+	      cur_y_offset == glyphs->glyphs[i].geometry.y_offset))
 	{
-	  if (glyphs->glyphs[i].glyph == 0)
+	  if (glyphs->glyphs[i].glyph == PANGO_GLYPH_EMPTY)
 	    {
-	      /* Code point 0 glyphs should not be rendered, but their
+	      /* PANGO_GLYPH_EMPTY glyphs should not be rendered, but their
 	       * indicated width (set up by PangoLayout) should be taken
 	       * into account.
 	       */
@@ -338,7 +339,7 @@ pango_win32_render (HDC               hdc,
 	    }
 	  else
 	    {
-	      if (glyphs->glyphs[i].glyph & PANGO_WIN32_UNKNOWN_FLAG)
+	      if (glyphs->glyphs[i].glyph & PANGO_GLYPH_UNKNOWN_FLAG)
 		{
 		  /* Glyph index is actually the char value that doesn't
 		   * have any glyph (ORed with the flag). We should really
@@ -371,12 +372,12 @@ pango_win32_render (HDC               hdc,
 #ifdef PANGO_WIN32_DEBUGGING
       if (pango_win32_debug)
 	{
-	  printf ("ExtTextOutW at %d,%d deltas:",
-		  x + PANGO_PIXELS (start_x_offset),
-		  y + PANGO_PIXELS (cur_y_offset));
+	  g_print ("ExtTextOutW at %d,%d deltas:",
+		   x + PANGO_PIXELS (start_x_offset),
+		   y + PANGO_PIXELS (cur_y_offset));
 	  for (j = 0; j < num_valid_glyphs; j++)
-	    printf (" %d", dX[j]);
-	  printf ("\n");
+	    g_print (" %d", dX[j]);
+	  g_print ("\n");
 	}
 #endif
 
@@ -396,6 +397,59 @@ pango_win32_render (HDC               hdc,
   g_free (dX);
 }
 
+/**
+ * pango_win32_render_transformed:
+ * @hdc:     a windows device context
+ * @matrix:  a #PangoMatrix, or %NULL to use an identity transformation
+ * @font:    the font in which to draw the string
+ * @glyphs:  the glyph string to draw
+ * @x:       the x position of the start of the string (in Pango
+ *           units in user space coordinates)
+ * @y:       the y position of the baseline (in Pango units
+ *           in user space coordinates)
+ *
+ * Renders a #PangoGlyphString onto a windows DC, possibly
+ * transforming the layed-out coordinates through a transformation
+ * matrix. Note that the transformation matrix for @font is not
+ * changed, so to produce correct rendering results, the @font
+ * must have been loaded using a #PangoContext with an identical
+ * transformation matrix to that passed in to this function.
+ **/
+void 
+pango_win32_render_transformed (HDC                hdc,
+				const PangoMatrix *matrix,
+				PangoFont         *font,
+				PangoGlyphString  *glyphs,
+				int                x, 
+				int                y)
+{
+  XFORM xForm;
+  XFORM xFormPrev = {1.0, 0.0, 0.0, 1.0, 0.0, 0.0};
+  int   mode = GetGraphicsMode (hdc);
+
+  if (!SetGraphicsMode (hdc, GM_ADVANCED))
+    g_warning ("SetGraphicsMode() failed");
+  else if (!GetWorldTransform (hdc, &xFormPrev))
+    g_warning ("GetWorldTransform() failed");
+  else if (matrix)
+    {
+      xForm.eM11 = matrix->xx;
+      xForm.eM12 = matrix->yx;
+      xForm.eM21 = matrix->xy;
+      xForm.eM22 = matrix->yy;
+      xForm.eDx = matrix->x0;
+      xForm.eDy = matrix->y0;
+      if (!SetWorldTransform (hdc, &xForm))
+        g_warning ("GetWorldTransform() failed");
+    }
+
+  pango_win32_render (hdc, font, glyphs, x/PANGO_SCALE, y/PANGO_SCALE);
+
+  /* restore */
+  SetWorldTransform (hdc, &xFormPrev);
+  SetGraphicsMode (hdc, mode);
+}
+
 static void
 pango_win32_font_get_glyph_extents (PangoFont      *font,
 				    PangoGlyph      glyph,
@@ -410,24 +464,23 @@ pango_win32_font_get_glyph_extents (PangoFont      *font,
   MAT2 m = {{0,1}, {0,0}, {0,0}, {0,1}};
   PangoWin32GlyphInfo *info;
 
-  if (glyph & PANGO_WIN32_UNKNOWN_FLAG)
+  if (glyph == PANGO_GLYPH_EMPTY)
+    {
+      if (ink_rect)
+	ink_rect->x = ink_rect->width = ink_rect->y = ink_rect->height = 0;
+      if (logical_rect)
+	logical_rect->x = logical_rect->width = logical_rect->y = logical_rect->height = 0;
+      return;
+    }
+
+  if (glyph & PANGO_GLYPH_UNKNOWN_FLAG)
     glyph_index = glyph = 0;
 
   info = g_hash_table_lookup (win32font->glyph_info, GUINT_TO_POINTER (glyph));
   
   if (!info)
     {
-      info = g_new (PangoWin32GlyphInfo, 1);
-
-      info->ink_rect.x = 0;
-      info->ink_rect.width = 0;
-      info->ink_rect.y = 0;
-      info->ink_rect.height = 0;
-
-      info->logical_rect.x = 0;
-      info->logical_rect.width = 0;
-      info->logical_rect.y = 0;
-      info->logical_rect.height = 0;
+      info = g_new0 (PangoWin32GlyphInfo, 1);
 
       memset (&gm, 0, sizeof (gm));
 
@@ -500,54 +553,122 @@ static PangoFontMetrics *
 pango_win32_font_get_metrics (PangoFont     *font,
 			      PangoLanguage *language)
 {
-  HFONT hfont;
-  TEXTMETRIC tm;
-  PangoLayout *layout;
-  PangoRectangle extents;
-  PangoContext *context;
-  PangoFontMetrics *metrics;
-
-  metrics = pango_font_metrics_new ();
+  PangoWin32MetricsInfo *info = NULL; /* Quiet gcc */
+  PangoWin32Font *win32font = (PangoWin32Font *)font;
+  GSList *tmp_list;
+      
+  const char *sample_str = pango_language_get_sample_string (language);
   
-  metrics->ascent = 0;
-  metrics->descent = 0;
-  metrics->approximate_digit_width = 0;
-  metrics->approximate_char_width = 0;
-  
-  hfont = pango_win32_get_hfont (font);
-
-  if (hfont != NULL)
+  tmp_list = win32font->metrics_by_lang;
+  while (tmp_list)
     {
-      PangoFontDescription *font_desc;
-	    
-      SelectObject (pango_win32_hdc, hfont);
-      GetTextMetrics (pango_win32_hdc, &tm);
+      info = tmp_list->data;
+      
+      if (info->sample_str == sample_str)    /* We _don't_ need strcmp */
+	break;
 
-      metrics->ascent = tm.tmAscent * PANGO_SCALE;
-      metrics->descent = tm.tmDescent * PANGO_SCALE;
-      metrics->approximate_char_width = tm.tmAveCharWidth * PANGO_SCALE;
-
-      /* lovely copy&paste programming (from pangox.c) FIXME: This is
-       * sort of a sledgehammer solution, if we cached the results,
-       * like the other backends did, it wouldn't big a huge performance
-       * problem.  Get the avg. width of the chars in "0123456789"
-       */
-      context = pango_win32_get_context ();
-      pango_context_set_language (context, language);
-      font_desc = pango_font_describe (font);
-      pango_context_set_font_description (context, font_desc);
-      layout = pango_layout_new (context);
-      pango_layout_set_text (layout, "0123456789", -1);
-      metrics->approximate_digit_width = max_glyph_width (layout);
-
-      pango_font_description_free (font_desc);
-      g_object_unref (layout);
-      g_object_unref (context);
+      tmp_list = tmp_list->next;
     }
 
-  return metrics;
+  if (!tmp_list)
+    {
+      HFONT hfont;
+      PangoFontMetrics *metrics;
+      
+      info = g_new (PangoWin32MetricsInfo, 1);
+      win32font->metrics_by_lang = g_slist_prepend (win32font->metrics_by_lang, info);
+
+      info->sample_str = sample_str;
+      info->metrics = metrics = pango_font_metrics_new ();
+
+      hfont = pango_win32_get_hfont (font);
+      if (hfont != NULL)
+	{
+	  PangoCoverage *coverage;
+	  TEXTMETRIC tm;
+	    
+	  SelectObject (pango_win32_hdc, hfont);
+	  GetTextMetrics (pango_win32_hdc, &tm);
+
+	  metrics->ascent = tm.tmAscent * PANGO_SCALE;
+	  metrics->descent = tm.tmDescent * PANGO_SCALE;
+	  metrics->approximate_char_width = tm.tmAveCharWidth * PANGO_SCALE;
+
+	  coverage = pango_win32_font_get_coverage (font, language);
+	  if (pango_coverage_get (coverage, '0') != PANGO_COVERAGE_NONE &&
+	      pango_coverage_get (coverage, '9') != PANGO_COVERAGE_NONE)
+	    {
+	      PangoContext *context;
+	      PangoFontDescription *font_desc;
+	      PangoLayout *layout;
+	      
+	      /*  Get the average width of the chars in "0123456789" */
+	      context = pango_win32_get_context ();
+	      pango_context_set_language (context, language);
+	      font_desc = pango_font_describe (font);
+	      pango_context_set_font_description (context, font_desc);
+	      layout = pango_layout_new (context);
+	      pango_layout_set_text (layout, "0123456789", -1);
+
+	      metrics->approximate_digit_width = max_glyph_width (layout);
+
+	      pango_font_description_free (font_desc);
+	      g_object_unref (layout);
+	      g_object_unref (context);
+	    }
+	  else
+	    metrics->approximate_digit_width = metrics->approximate_char_width;
+
+	  pango_coverage_unref (coverage);
+
+	  /* FIXME: Should get the real values from the TrueType font file */
+	  metrics->underline_position = -2 * PANGO_SCALE;
+	  metrics->underline_thickness = 1 * PANGO_SCALE;
+	  metrics->strikethrough_thickness = metrics->underline_thickness;
+	  /* Really really wild guess */
+	  metrics->strikethrough_position = metrics->ascent / 3;
+	}
+    }
+      
+  return pango_font_metrics_ref (info->metrics);
 }
 
+static PangoFontMap *
+pango_win32_font_get_font_map (PangoFont *font)
+{
+  PangoWin32Font *win32font = (PangoWin32Font *)font;
+
+  return win32font->fontmap;
+}
+
+static gboolean
+pango_win32_font_real_select_font (PangoFont *font,
+				   HDC        hdc)
+{
+  HFONT hfont = pango_win32_get_hfont (font);
+
+  if (!hfont)
+    return FALSE;
+  
+  if (!SelectObject (hdc, hfont))
+    {
+      g_warning ("pango_win32_font_real_select_font: Cannot select font\n");
+      return FALSE;
+    }
+  
+  return TRUE;
+}
+
+static void
+pango_win32_font_real_done_font (PangoFont *font)
+{
+}
+
+static double
+pango_win32_font_real_get_metrics_factor (PangoFont *font)
+{
+  return PANGO_SCALE;
+}
 
 /**
  * pango_win32_font_logfont:
@@ -573,6 +694,65 @@ pango_win32_font_logfont (PangoFont *font)
   return lfp;
 }
 
+/**
+ * pango_win32_font_select_font:
+ * @font: a #PangoFont from the Win32 backend
+ * @hdc: a windows device context
+ * 
+ * Selects the font into the specified DC and changes the mapping mode
+ * and world transformation of the DC appropriately for the font.
+ * You may want to surround the use of this function with calls
+ * to SaveDC and RestoreDC. Call pango_win32_font_done_font() when
+ * you are done using the DC to release allocated resources.
+ *
+ * See pango_win32_font_get_metrics_factor() for information about
+ * converting from the coordinate space used by this function
+ * into Pango units.
+ * 
+ * Return value: %TRUE if the operation succeeded.
+ **/
+gboolean
+pango_win32_font_select_font (PangoFont *font,
+			      HDC        hdc)
+{
+  g_return_val_if_fail (PANGO_WIN32_IS_FONT (font), FALSE);
+  
+  return PANGO_WIN32_FONT_GET_CLASS (font)->select_font (font, hdc);
+}
+
+/**
+ * pango_win32_font_done_font:
+ * @font: a #PangoFont from the win32 backend
+ *
+ * Releases any resources allocated by pango_win32_font_done_font()
+ **/
+void
+pango_win32_font_done_font (PangoFont *font)
+{
+  g_return_if_fail (PANGO_WIN32_IS_FONT (font));
+  
+  PANGO_WIN32_FONT_GET_CLASS (font)->done_font (font);
+}
+
+/**
+ * pango_win32_font_get_metrics_factor:
+ * @font: a #PangoFont from the win32 backend
+ * 
+ * Returns the scale factor from logical units in the coordinate
+ * space used by pango_win32_font_select_font() to Pango units
+ * in user space.
+ * 
+ * Return value: factor to multiply logical units by to get Pango
+ *               units.
+ **/
+double
+pango_win32_font_get_metrics_factor (PangoFont *font)
+{
+  g_return_val_if_fail (PANGO_WIN32_IS_FONT (font), 1.);
+  
+  return PANGO_WIN32_FONT_GET_CLASS (font)->get_metrics_factor (font);
+}
+
 static void
 pango_win32_font_dispose (GObject *object)
 {
@@ -588,6 +768,13 @@ pango_win32_font_dispose (GObject *object)
   G_OBJECT_CLASS (pango_win32_font_parent_class)->dispose (object);
 }
 
+static void
+free_metrics_info (PangoWin32MetricsInfo *info)
+{
+  pango_font_metrics_unref (info->metrics);
+  g_free (info);
+}
+
 static void
 pango_win32_font_finalize (GObject *object)
 {
@@ -597,6 +784,9 @@ pango_win32_font_finalize (GObject *object)
   if (win32font->hfont != NULL)
     pango_win32_font_cache_unload (cache, win32font->hfont);
 
+  g_slist_foreach (win32font->metrics_by_lang, (GFunc)free_metrics_info, NULL);
+  g_slist_free (win32font->metrics_by_lang);
+  
   if (win32font->win32face)
     pango_win32_font_entry_remove (win32font->win32face, PANGO_FONT (win32font));
  
@@ -673,6 +863,8 @@ pango_win32_font_find_shaper (PangoFont     *font,
  * 
  * Returns the index of a glyph suitable for drawing @wc as an
  * unknown character.
+ *
+ * Use PANGO_GET_UNKNOWN_GLYPH() instead.
  * 
  * Return value: a glyph index into @font
  **/
@@ -680,17 +872,19 @@ PangoGlyph
 pango_win32_get_unknown_glyph (PangoFont *font,
 			       gunichar   wc)
 {
-  return wc | PANGO_WIN32_UNKNOWN_FLAG;
+  return PANGO_GET_UNKNOWN_GLYPH (wc);
 }
 
 /**
  * pango_win32_render_layout_line:
- * @hdc:       HDC to use for uncolored drawing
+ * @hdc:       DC to use for uncolored drawing
  * @line:      a #PangoLayoutLine
  * @x:         the x position of start of string (in pixels)
  * @y:         the y position of baseline (in pixels)
  *
- * Render a #PangoLayoutLine onto a device context
+ * Render a #PangoLayoutLine onto a device context. For underlining to
+ * work property the text alignment of the DC should have TA_BASELINE
+ * and TA_LEFT.
  */
 void 
 pango_win32_render_layout_line (HDC              hdc,
@@ -764,11 +958,15 @@ pango_win32_render_layout_line (HDC              hdc,
 	  points[0].y = points[1].y = y + 4;
 	  points[1].x = x + PANGO_PIXELS (x_off + ink_rect.x + ink_rect.width);
 	  Polyline (hdc, points, 2);
-	  /* Fall through */
-	case PANGO_UNDERLINE_SINGLE:
 	  points[0].y = points[1].y = y + 2;
 	  Polyline (hdc, points, 2);
 	  break;
+	case PANGO_UNDERLINE_SINGLE:
+	  points[0].x = x + PANGO_PIXELS (x_off + ink_rect.x) - 1;
+	  points[0].y = points[1].y = y + 2;
+	  points[1].x = x + PANGO_PIXELS (x_off + ink_rect.x + ink_rect.width);
+	  Polyline (hdc, points, 2);
+	  break;
 	case PANGO_UNDERLINE_ERROR:
           {
             int point_x;
@@ -907,7 +1105,8 @@ pango_win32_get_item_properties (PangoItem      *item,
 }
 
 static guint 
-get_unicode_mapping_offset (HDC hdc)
+get_cmap_offset (HDC     hdc,
+		 guint16 encoding_id)
 {
   guint16 n_tables;
   struct cmap_encoding_subtable *table;
@@ -916,8 +1115,8 @@ get_unicode_mapping_offset (HDC hdc)
   guint32 offset;
 
   /* Get The number of encoding tables, at offset 2 */
-  res = GetFontData (hdc, CMAP, 2, &n_tables, sizeof (guint16));
-  if (res != sizeof (guint16))
+  res = GetFontData (hdc, CMAP, 2, &n_tables, 2);
+  if (res != 2)
     return 0;
 
   n_tables = GUINT16_FROM_BE (n_tables);
@@ -931,7 +1130,7 @@ get_unicode_mapping_offset (HDC hdc)
   for (i = 0; i < n_tables; i++)
     {
       if (table[i].platform_id == GUINT16_TO_BE (MICROSOFT_PLATFORM_ID) &&
-	  table[i].encoding_id == GUINT16_TO_BE (UNICODE_ENCODING_ID))
+	  table[i].encoding_id == GUINT16_TO_BE (encoding_id))
 	{
 	  offset = GUINT32_FROM_BE (table[i].offset);
 	  g_free (table);
@@ -942,31 +1141,33 @@ get_unicode_mapping_offset (HDC hdc)
   return 0;
 }
 
-static struct type_4_cmap *
-get_unicode_mapping (HDC hdc)
+static gpointer
+get_format_4_cmap (HDC hdc)
 {
   guint32 offset;
   guint32 res;
   guint16 length;
   guint16 *tbl, *tbl_end;
-  struct type_4_cmap *table;
+  struct format_4_cmap *table;
 
   /* FIXME: Could look here at the CRC for the font in the DC 
             and return a cached copy if the same */
 
-  offset = get_unicode_mapping_offset (hdc);
+  offset = get_cmap_offset (hdc, UNICODE_ENCODING_ID);
   if (offset == 0)
     return NULL;
 
-  res = GetFontData (hdc, CMAP, offset + 2, &length, sizeof (guint16));
-  if (res != sizeof (guint16))
+  res = GetFontData (hdc, CMAP, offset + 2, &length, 2);
+  if (res != 2)
     return NULL;
   length = GUINT16_FROM_BE (length);
 
   table = g_malloc (length);
 
   res = GetFontData (hdc, CMAP, offset, table, length);
-  if (res != length)
+  if (res != length ||
+      GUINT16_FROM_BE (table->format) != 4 ||
+      (GUINT16_FROM_BE (table->length) % 2) != 0)
     {
       g_free (table);
       return NULL;
@@ -980,12 +1181,6 @@ get_unicode_mapping (HDC hdc)
   table->entry_selector = GUINT16_FROM_BE (table->entry_selector);
   table->range_shift = GUINT16_FROM_BE (table->range_shift);
 
-  if (table->format != 4)
-    {
-      g_free (table);
-      return NULL;
-    }
-  
   tbl_end = (guint16 *)((char *)table + length);
   tbl = &table->reserved;
 
@@ -999,28 +1194,28 @@ get_unicode_mapping (HDC hdc)
 }
 
 static guint16 *
-get_id_range_offset (struct type_4_cmap *table)
+get_id_range_offset (struct format_4_cmap *table)
 {
   gint32 seg_count = table->seg_count_x_2/2;
   return &table->arrays[seg_count*3];
 }
 
 static guint16 *
-get_id_delta (struct type_4_cmap *table)
+get_id_delta (struct format_4_cmap *table)
 {
   gint32 seg_count = table->seg_count_x_2/2;
   return &table->arrays[seg_count*2];
 }
 
 static guint16 *
-get_start_count (struct type_4_cmap *table)
+get_start_count (struct format_4_cmap *table)
 {
   gint32 seg_count = table->seg_count_x_2/2;
   return &table->arrays[seg_count*1];
 }
 
 static guint16 *
-get_end_count (struct type_4_cmap *table)
+get_end_count (struct format_4_cmap *table)
 {
   gint32 seg_count = table->seg_count_x_2/2;
   /* Apparently the reseved spot is not reserved for 
@@ -1028,11 +1223,10 @@ get_end_count (struct type_4_cmap *table)
   return (&table->arrays[seg_count*0])-1;
 }
 
-
 static gboolean
-find_segment (struct type_4_cmap *table, 
-	      guint16             wc,
-	      guint16            *segment)
+find_segment (struct format_4_cmap *table, 
+	      guint16               wc,
+	      guint16              *segment)
 {
   guint16 start, end, i;
   guint16 seg_count = table->seg_count_x_2/2;
@@ -1090,72 +1284,162 @@ find_segment (struct type_4_cmap *table,
   return FALSE;
 }
 
-static struct type_4_cmap *
-font_get_unicode_table (PangoFont *font)
+static gpointer
+get_format_12_cmap (HDC hdc)
 {
-  PangoWin32Font *win32font = (PangoWin32Font *)font;
-  HFONT hfont;
-  struct type_4_cmap *table;
-
-  if (win32font->win32face->unicode_table)
-    return (struct type_4_cmap *)win32font->win32face->unicode_table;
-
-  hfont = pango_win32_get_hfont (font);
-  SelectObject (pango_win32_hdc, hfont);
-  table = get_unicode_mapping (pango_win32_hdc);
+  guint32 offset;
+  guint32 res;
+  guint32 length;
+  guint32 *tbl, *tbl_end;
+  struct format_12_cmap *table;
   
-  win32font->win32face->unicode_table = table;
+  offset = get_cmap_offset (hdc, UCS4_ENCODING_ID);
+  if (offset == 0)
+    return NULL;
+
+  res = GetFontData (hdc, CMAP, offset + 4, &length, 4);
+  if (res != 4)
+    return NULL;
+  length = GUINT32_FROM_BE (length);
+
+  table = g_malloc (length);
+
+  res = GetFontData (hdc, CMAP, offset, table, length);
+  if (res != length)
+    {
+      g_free (table);
+      return NULL;
+    }
+
+  table->format = GUINT16_FROM_BE (table->format);
+  table->length = GUINT32_FROM_BE (table->length);
+  table->language = GUINT32_FROM_BE (table->language);
+  table->count = GUINT32_FROM_BE (table->count);
+
+  if (table->format != 12 ||
+      (table->length % 4) != 0 ||
+      table->length > length ||
+      table->length < 16 + table->count * 12)
+    {
+      g_free (table);
+      return NULL;
+    }
+  
+  tbl_end = (guint32 *) ((char *) table + length);
+  tbl = table->groups;
+
+  while (tbl < tbl_end)
+    {
+      *tbl = GUINT32_FROM_BE (*tbl);
+      tbl++;
+    }
 
   return table;
 }
+  
+static gpointer
+font_get_cmap (PangoFont *font)
+{
+  PangoWin32Font *win32font = (PangoWin32Font *)font;
+  gpointer cmap;
+
+  if (win32font->win32face->cmap)
+    return win32font->win32face->cmap;
+
+  pango_win32_font_select_font (font, pango_win32_hdc);
+
+  /* Prefer the format 12 cmap */
+  if ((cmap = get_format_12_cmap (pango_win32_hdc)) != NULL)
+    {
+      win32font->win32face->cmap_format = 12;
+      win32font->win32face->cmap = cmap;
+    }
+  else if ((cmap = get_format_4_cmap (pango_win32_hdc)) != NULL)
+    {
+      win32font->win32face->cmap_format = 4;
+      win32font->win32face->cmap = cmap;
+    }
+
+  pango_win32_font_done_font (font);
+
+  return cmap;
+}
 
 /**
  * pango_win32_font_get_glyph_index:
  * @font: a #PangoFont.
  * @wc: a Unicode character.
  *
- * Obtains the index of the glyph for @wc in @font.
+ * Obtains the index of the glyph for @wc in @font, or 0, if not
+ * covered.
  *
- * Returns: the glyph index for @wc.
+ * Return value: the glyph index for @wc.
  **/
 gint
 pango_win32_font_get_glyph_index (PangoFont *font,
 				  gunichar   wc)
 {
-  struct type_4_cmap *table;
-  guint16 *id_range_offset;
-  guint16 *id_delta;
-  guint16 *start_count;
-  guint16 segment;
-  guint16 id;
-  guint16 ch = wc;
+  PangoWin32Font *win32font = (PangoWin32Font *)font;
+  gpointer cmap;
   guint16 glyph;
 
   /* Do GetFontData magic on font->hfont here. */
-  table = font_get_unicode_table (font);
+  cmap = font_get_cmap (font);
 
-  if (table == NULL)
-    return 0;
-  
-  if (!find_segment (table, ch, &segment))
+  if (cmap == NULL)
     return 0;
 
-  id_range_offset = get_id_range_offset (table);
-  id_delta = get_id_delta (table);
-  start_count = get_start_count (table);
-
-  if (id_range_offset[segment] == 0)
-    glyph = (id_delta[segment] + ch) % 65536;
-  else
+  if (win32font->win32face->cmap_format == 4)
     {
-      id = *(id_range_offset[segment]/2 +
-	     (ch - start_count[segment]) +
-	     &id_range_offset[segment]);
-      if (id)
-	glyph = (id_delta[segment] + id) %65536;
+      struct format_4_cmap *cmap4 = cmap;
+      guint16 *id_range_offset;
+      guint16 *id_delta;
+      guint16 *start_count;
+      guint16 segment;
+      guint16 id;
+      guint16 ch = wc;
+  
+      if (wc > 0xFFFF)
+	return 0;
+
+      if (!find_segment (cmap4, ch, &segment))
+	return 0;
+
+      id_range_offset = get_id_range_offset (cmap4);
+      id_delta = get_id_delta (cmap4);
+      start_count = get_start_count (cmap4);
+
+      if (id_range_offset[segment] == 0)
+	glyph = (id_delta[segment] + ch) % 65536;
       else
-	glyph = 0;
+	{
+	  id = *(id_range_offset[segment]/2 +
+		 (ch - start_count[segment]) +
+		 &id_range_offset[segment]);
+	  if (id)
+	    glyph = (id_delta[segment] + id) %65536;
+	  else
+	    glyph = 0;
+	}
     }
+  else if (win32font->win32face->cmap_format == 12)
+    {
+      struct format_12_cmap *cmap12 = cmap;
+      guint32 i;
+
+      glyph = 0;
+      for (i = 0; i < cmap12->count; i++)
+	{
+	  if (cmap12->groups[i*3+0] <= wc && wc <= cmap12->groups[i*3+1])
+	    {
+	      glyph = cmap12->groups[i*3+2] + (wc - cmap12->groups[i*3+0]);
+	      break;
+	    }
+	}
+    }
+  else
+    g_assert_not_reached ();
+
   return glyph;
 }
 
@@ -1258,14 +1542,10 @@ pango_win32_font_calc_coverage (PangoFont     *font,
 				PangoCoverage *coverage,
 				PangoLanguage *lang)
 {
-  struct type_4_cmap *table;
-  guint16 *id_range_offset;
-  guint16 *start_count;
-  guint16 *end_count;
-  guint16 seg_count;
-  guint16 id;
+  PangoWin32Font *win32font = (PangoWin32Font *)font;
+  gpointer cmap;
   guint32 ch;
-  int i;
+  guint32 i;
   PangoWin32CoverageLanguageClass cjkv;
   gboolean hide_unihan = FALSE;
   
@@ -1282,98 +1562,128 @@ pango_win32_font_calc_coverage (PangoFont     *font,
     }
 
   /* Do GetFontData magic on font->hfont here. */
-
-  table = font_get_unicode_table (font);
-
-  if (table == NULL)
+  cmap = font_get_cmap (font);
+  if (cmap == NULL)
     {
       PING(("no table"));
       return;
     }
   
-  seg_count = table->seg_count_x_2/2;
-  end_count = get_end_count (table);
-  start_count = get_start_count (table);
-  id_range_offset = get_id_range_offset (table);
-
   PING (("coverage:"));
-  for (i = 0; i < seg_count; i++)
+  if (win32font->win32face->cmap_format == 4)
     {
-      if (id_range_offset[i] == 0)
-	{
-#ifdef PANGO_WIN32_DEBUGGING
-	  if (pango_win32_debug)
-	    {
-	      if (end_count[i] == start_count[i])
-		printf ("%04x ", start_count[i]);
-	      else
-		printf ("%04x:%04x ", start_count[i], end_count[i]);
-	    }
-#endif	  
-	  for (ch = start_count[i]; 
-	       ch <= end_count[i];
-	       ch++)
-	    if (hide_unihan && ch >= 0x3400 && ch <= 0x9FAF)
-	      pango_coverage_set (coverage, ch, PANGO_COVERAGE_APPROXIMATE);
-	    else
-	      pango_coverage_set (coverage, ch, PANGO_COVERAGE_EXACT);
-	}
-      else
-	{
-#ifdef PANGO_WIN32_DEBUGGING
-	  guint32 ch0 = G_MAXUINT;
-#endif
-	  for (ch = start_count[i]; 
-	       ch <= end_count[i];
-	       ch++)
-	    {
-	      if (ch == 0xFFFF)
-		break;
+      struct format_4_cmap *cmap4 = cmap;
+      guint16 *id_range_offset;
+      guint16 *start_count;
+      guint16 *end_count;
+      guint16 seg_count;
+      guint16 id;
 
-	      id = *(id_range_offset[i]/2 +
-		     (ch - start_count[i]) +
-		     &id_range_offset[i]);
-	      if (id)
-		{
+      seg_count = cmap4->seg_count_x_2/2;
+      end_count = get_end_count (cmap4);
+      start_count = get_start_count (cmap4);
+      id_range_offset = get_id_range_offset (cmap4);
+
+      for (i = 0; i < seg_count; i++)
+	{
+	  if (id_range_offset[i] == 0)
+	    {
 #ifdef PANGO_WIN32_DEBUGGING
-		  if (ch0 == G_MAXUINT)
-		    ch0 = ch;
-#endif
-		  if (hide_unihan && ch >= 0x3400 && ch <= 0x9FAF)
-		    pango_coverage_set (coverage, ch, PANGO_COVERAGE_APPROXIMATE);
+	      if (pango_win32_debug)
+		{
+		  if (end_count[i] == start_count[i])
+		    g_print ("%04x ", start_count[i]);
 		  else
-		    pango_coverage_set (coverage, ch, PANGO_COVERAGE_EXACT);
+		    g_print ("%04x:%04x ", start_count[i], end_count[i]);
+		}
+#endif	  
+	      for (ch = start_count[i]; ch <= end_count[i]; ch++)
+		if (hide_unihan && CH_IS_UNIHAN_BMP (ch))
+		  pango_coverage_set (coverage, ch, PANGO_COVERAGE_APPROXIMATE);
+		else
+		  pango_coverage_set (coverage, ch, PANGO_COVERAGE_EXACT);
+	    }
+	  else
+	    {
+#ifdef PANGO_WIN32_DEBUGGING
+	      guint32 ch0 = G_MAXUINT;
+#endif
+	      for (ch = start_count[i]; ch <= end_count[i]; ch++)
+		{
+		  if (ch == 0xFFFF)
+		    break;
+		  
+		  id = *(id_range_offset[i]/2 +
+			 (ch - start_count[i]) +
+			 &id_range_offset[i]);
+		  if (id)
+		    {
+#ifdef PANGO_WIN32_DEBUGGING
+		      if (ch0 == G_MAXUINT)
+			ch0 = ch;
+#endif
+		      if (hide_unihan && CH_IS_UNIHAN_BMP (ch))
+			pango_coverage_set (coverage, ch, PANGO_COVERAGE_APPROXIMATE);
+		      else
+			pango_coverage_set (coverage, ch, PANGO_COVERAGE_EXACT);
+		    }
+#ifdef PANGO_WIN32_DEBUGGING
+		  else if (ch0 < G_MAXUINT)
+		    {
+		      if (pango_win32_debug)
+			{
+			  if (ch > ch0 + 2)
+			    g_print ("%04x:%04x ", ch0, ch - 1);
+			  else
+			    g_print ("%04x ", ch0);
+			}
+		      ch0 = G_MAXUINT;
+		    }
+#endif
 		}
 #ifdef PANGO_WIN32_DEBUGGING
-	      else if (ch0 < G_MAXUINT)
+	      if (ch0 < G_MAXUINT)
 		{
 		  if (pango_win32_debug)
 		    {
 		      if (ch > ch0 + 2)
-			printf ("%04x:%04x ", ch0, ch - 1);
+			g_print ("%04x:%04x ", ch0, ch - 1);
 		      else
-			printf ("%04x ", ch0);
+			g_print ("%04x ", ch0);
 		    }
-		  ch0 = G_MAXUINT;
 		}
 #endif
 	    }
-#ifdef PANGO_WIN32_DEBUGGING
-	  if (ch0 < G_MAXUINT)
-	    {
-	      if (pango_win32_debug)
-		{
-		  if (ch > ch0 + 2)
-		    printf ("%04x:%04x ", ch0, ch - 1);
-		  else
-		    printf ("%04x ", ch0);
-		}
-	    }
-#endif
 	}
     }
+  else if (win32font->win32face->cmap_format == 12)
+    {
+      struct format_12_cmap *cmap12 = cmap;
+
+      for (i = 0; i < cmap12->count; i++)
+	{
+#ifdef PANGO_WIN32_DEBUGGING
+	  if (pango_win32_debug)
+	    {
+	      if (cmap12->groups[i*3+0] == cmap12->groups[i*3+1])
+		g_print ("%04x ", cmap12->groups[i*3+0]);
+	      else
+		g_print ("%04x:%04x ", cmap12->groups[i*3+0], cmap12->groups[i*3+1]);
+	    }
+#endif
+	  for (ch = cmap12->groups[i*3+0]; ch <= cmap12->groups[i*3+1]; ch++)
+	    {
+	      if (hide_unihan && CH_IS_UNIHAN (ch))
+		pango_coverage_set (coverage, ch, PANGO_COVERAGE_APPROXIMATE);
+	      else
+		pango_coverage_set (coverage, ch, PANGO_COVERAGE_EXACT);
+	    }
+	}
+    }
+  else
+    g_assert_not_reached ();
 #ifdef PANGO_WIN32_DEBUGGING
   if (pango_win32_debug)
-    printf ("\n");
+    g_print ("\n");
 #endif
 }
